@@ -24,9 +24,9 @@ use tokio::{net::UdpSocket, sync::Mutex as AsyncMutex};
 use tokio_util::sync::CancellationToken;
 use zenoh_core::{zasynclock, zlock};
 use zenoh_link_commons::{
-    get_ip_interface_names, ConstructibleLinkManagerUnicast, LinkAuthId, LinkManagerUnicastTrait,
-    LinkUnicast, LinkUnicastTrait, ListenersUnicastIP, NewLinkChannelSender, BIND_INTERFACE,
-    BIND_SOCKET,
+    get_ip_interface_names, parse_dscp, set_dscp, ConstructibleLinkManagerUnicast, LinkAuthId,
+    LinkManagerUnicastTrait, LinkUnicast, LinkUnicastTrait, ListenersUnicastIP,
+    NewLinkChannelSender, BIND_INTERFACE, BIND_SOCKET,
 };
 use zenoh_protocol::{
     core::{Address, EndPoint, Locator},
@@ -39,6 +39,7 @@ use super::{
     get_udp_addrs, socket_addr_to_udp_locator, UDP_ACCEPT_THROTTLE_TIME, UDP_DEFAULT_MTU,
     UDP_MAX_MTU,
 };
+use crate::pktinfo;
 
 type LinkHashMap = Arc<Mutex<HashMap<(SocketAddr, SocketAddr), Weak<LinkUnicastUdpUnconnected>>>>;
 type LinkInput = (Vec<u8>, usize);
@@ -273,6 +274,7 @@ impl LinkManagerUnicastUdp {
         dst_addr: &SocketAddr,
         iface: Option<&str>,
         bind_socket: Option<&str>,
+        dscp: Option<u32>,
     ) -> ZResult<(UdpSocket, SocketAddr, SocketAddr)> {
         let src_socket_addr = if let Some(bind_socket) = bind_socket {
             let address = Address::from(bind_socket);
@@ -301,6 +303,10 @@ impl LinkManagerUnicastUdp {
             zenoh_util::net::set_bind_to_device_udp_socket(&socket, iface)?;
         }
 
+        if let Some(dscp) = dscp {
+            set_dscp(&socket, *dst_addr, dscp)?;
+        }
+
         // Connect the socket to the remote address
         socket.connect(dst_addr).await.map_err(|e| {
             let e = zerror!("Can not connect a new UDP link to {}: {}", dst_addr, e);
@@ -338,6 +344,7 @@ impl LinkManagerUnicastUdp {
         &self,
         addr: &SocketAddr,
         iface: Option<&str>,
+        dscp: Option<u32>,
     ) -> ZResult<(UdpSocket, SocketAddr)> {
         // Bind the UDP socket
         let socket = UdpSocket::bind(addr).await.map_err(|e| {
@@ -350,6 +357,10 @@ impl LinkManagerUnicastUdp {
             zenoh_util::net::set_bind_to_device_udp_socket(&socket, iface)?;
         }
 
+        if let Some(dscp) = dscp {
+            set_dscp(&socket, *addr, dscp)?;
+        }
+
         let local_addr = socket.local_addr().map_err(|e| {
             let e = zerror!("Can not create a new UDP listener on {}: {}", addr, e);
             tracing::warn!("{}", e);
@@ -378,10 +389,11 @@ impl LinkManagerUnicastTrait for LinkManagerUnicastUdp {
         }
 
         let bind_socket = config.get(BIND_SOCKET);
+        let dscp = parse_dscp(&config)?;
 
         let mut errs: Vec<ZError> = vec![];
         for da in dst_addrs {
-            match self.new_link_inner(&da, iface, bind_socket).await {
+            match self.new_link_inner(&da, iface, bind_socket, dscp).await {
                 Ok((socket, src_addr, dst_addr)) => {
                     // Create UDP link
                     let link = Arc::new(LinkUnicastUdp::new(
@@ -417,10 +429,11 @@ impl LinkManagerUnicastTrait for LinkManagerUnicastUdp {
             .filter(|a| !a.ip().is_multicast());
         let config = endpoint.config();
         let iface = config.get(BIND_INTERFACE);
+        let dscp = parse_dscp(&config)?;
 
         let mut errs: Vec<ZError> = vec![];
         for da in addrs {
-            match self.new_listener_inner(&da, iface).await {
+            match self.new_listener_inner(&da, iface, dscp).await {
                 Ok((socket, local_addr)) => {
                     // Update the endpoint locator address
                     endpoint = EndPoint::new(
@@ -528,22 +541,26 @@ async fn accept_read_task(
         };
     }
 
-    async fn receive(socket: Arc<UdpSocket>, buffer: &mut [u8]) -> ZResult<(usize, SocketAddr)> {
-        let res = socket.recv_from(buffer).await.map_err(|e| zerror!(e))?;
+    async fn receive(
+        socket: Arc<pktinfo::PktInfoUdpSocket>,
+        buffer: &mut [u8],
+    ) -> ZResult<(usize, SocketAddr, SocketAddr)> {
+        let res = socket.receive(buffer).await.map_err(|e| zerror!(e))?;
         Ok(res)
     }
 
-    let src_addr = socket.local_addr().map_err(|e| {
+    let local_src_addr = socket.local_addr().map_err(|e| {
         let e = zerror!("Can not accept UDP connections: {}", e);
         tracing::warn!("{}", e);
         e
     })?;
+    let socket = Arc::new(pktinfo::PktInfoUdpSocket::new(socket).map_err(|e| {
+        let e = zerror!("Failed to enable IP_PKTINFO: {}", e);
+        tracing::warn!("{}", e);
+        e
+    })?);
 
-    tracing::trace!("Ready to accept UDP connections on: {:?}", src_addr);
-
-    if src_addr.ip().is_unspecified() {
-        tracing::warn!("Interceptors (e.g. Access Control, Downsampling) are not guaranteed to work on UDP when listening on 0.0.0.0 or [::]. Their usage is discouraged. See https://github.com/eclipse-zenoh/zenoh/issues/1126.");
-    }
+    tracing::trace!("Ready to accept UDP connections on: {:?}", local_src_addr);
 
     loop {
         // Buffers for deserialization
@@ -554,7 +571,7 @@ async fn accept_read_task(
 
             res = receive(socket.clone(), &mut buff) => {
                 match res {
-                    Ok((n, dst_addr)) => {
+                    Ok((n, dst_addr, src_addr)) => {
                         let link = loop {
                             let res = zgetlink!(src_addr, dst_addr);
                             match res {
@@ -563,7 +580,7 @@ async fn accept_read_task(
                                     // A new peers has sent data to this socket
                                     tracing::debug!("Accepted UDP connection on {}: {}", src_addr, dst_addr);
                                     let unconnected = Arc::new(LinkUnicastUdpUnconnected {
-                                        socket: Arc::downgrade(&socket),
+                                        socket: Arc::downgrade(&socket.socket),
                                         links: links.clone(),
                                         input: Mvar::new(),
                                         leftover: AsyncMutex::new(None),
